==> ocp-0.1.0.tar.bz2: ocp-0.1.0/stuff/timer.c <==
/* OpenCP Module Player
 * copyright (c) '94-'04 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * Systemtimer handlers
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -doj990328  Dirk Jagdmann  <doj@cubic.org>
 *    -changed interrupt access calls to calls from irq.h
 *  -fd990518   Felix Domke  <tmbinc@gmx.net>
 *    -added CLD after the tmOldTimer-call.
 *     this removed the devwmix*-STRANGEBUG. (finally, i hope)
 *  -fd990817   Felix Domke  <tmbinc@gmx.net>
 *    -added tmSetSecure/tmReleaseSecure to ensure that timer is only
 *     called when not "indos". needed for devpVXD (and some other maybe).
 *  -ss040613   Stian Skjelstad <stian@nixia.no>
 *    -Rewritten to use the posix itimer, and posix signals to fetch them
 *  --ss040907  Stian Skjelstad <stian@nixia.no>
 *    -Use gettimeofday() to calculate cpu-usage, since itimer() uses rounded off values
 */

#include "config.h"
#include <signal.h>
#include <stdio.h>
#include <sys/time.h>
#include "irq.h"
#include "imsrtns.h"
#include "timer.h"
#include "poll.h"

#define CLOCK_TICK_RATE 1193180 /* Base hertz of the i8253 or what ever that chip was namned - Stian */

static unsigned long count_to_time(unsigned long ticks)
{
	unsigned long retval;

	if (ticks&0xfffff000)
	{
		ticks*=62500; /* this is safe, since counter on the chip is 16 bit */
		retval = ticks / CLOCK_TICK_RATE;
		retval*=16;
	} else {
		ticks*=1000000;
		retval = ticks / CLOCK_TICK_RATE;
	}
	return retval;
}

static unsigned long time_to_count(unsigned long time)
{
	unsigned long retval;
	
	if (time<=4000)
	{
		retval = time * CLOCK_TICK_RATE / 1000000;
	} else {
		time/=4000;
		retval = time * CLOCK_TICK_RATE / 2500;
	}
	return retval;
}

static void (*tmTimerRoutine)()=NULL;
static void (*tmTimerRoutineSlave)()=NULL;

static unsigned long tmTimerRate;
static unsigned long tmIntCount;
static unsigned long tmTicker;
#ifdef TIMER_DEBUG
static int tmInited = 0;
#endif

static volatile char overload;
static volatile float cpuusage;
static int secure=0;

static int stackused; /* we don't provide a stack, but is for locking */

static void tmTimerHandler(int ignore)
{
#if 0
	char fpubuf[108]; /* 108 is enough acording to my intel book */
	__asm__ __volatile__ ("fsave %0\n" : : "m"(*fpubuf));
#endif

	struct timeval pre, pro;

	gettimeofday(&pre, NULL);
	tmTicker+=tmTimerRate;

	tmIntCount+=tmTimerRate;
	if (tmIntCount&0xFFFFc000)
	{
		tmIntCount&=0x3FFF;
		if (tmTimerRoutineSlave)
			tmTimerRoutineSlave();
	}

	if (stackused)
	{
		cpuusage=100;
		overload=1;
		return;
	}

	stackused++;
	if (!secure)
	{
		if (tmTimerRoutine)
			tmTimerRoutine();
	}
	stackused--;

	if (!overload)
	{
		struct itimerval spec;
		unsigned long spent;

		getitimer(ITIMER_REAL, &spec);
		gettimeofday(&pro, NULL);
		spent=(pro.tv_sec-pre.tv_sec)*1000000+pro.tv_usec-pre.tv_usec;
		cpuusage=0.1*(100.0*spent/spec.it_interval.tv_usec)+0.9*cpuusage;
	} else
		cpuusage=100;
	overload=0;
#if 0
	__asm__ __volatile__ ("frstor %0\n" : : "m"(*fpubuf));
#endif
}

int tmInit(void (*rout)(), int timerval)
{
	struct itimerval spec;

#ifdef TIMER_DEBUG
	if (tmInited)
	{
		fprintf(stderr, "tmInit: we are already inited\n");
	} else {
		tmInited=1;
	}
#endif

	tmTimerRate = timerval;
	tmTicker = -timerval;
#ifdef TIMER_DEBUG
	fprintf(stderr, "tmInit (ticks=%d, time=%ld)\n", timerval, count_to_time(timerval));
#endif
	timerval = count_to_time(timerval);
  
	tmTimerRoutine=rout;
	tmIntCount=0;

	irqInit(SIGALRM, tmTimerHandler, 0 /* glibc is broken, so don't set this to true*/);
	spec.it_interval.tv_sec=0;
	spec.it_interval.tv_usec=timerval;
	spec.it_value.tv_sec=0;
	spec.it_value.tv_usec=timerval;
	setitimer(ITIMER_REAL, &spec, NULL);

	cpuusage=0;

	return 1;
}

void tmResetTimer(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmResetTimer: we are not inited\n");
#endif
	if (!(tmTimerRoutine&&tmTimerRoutineSlave))
	{
		/* shut down irq */
		struct itimerval spec;

		spec.it_interval.tv_sec=0;
		spec.it_interval.tv_usec=0;
		spec.it_value.tv_sec=0;
		spec.it_value.tv_usec=0;
		setitimer(ITIMER_REAL, &spec, NULL);
 
		irqDone(SIGALRM);
	} else if (!tmTimerRoutine)
		/* recalc to a less/more aggressive clock */
		tmSetNewRate(17100);

}

void tmClose(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmClose: we are not inited\n");
#endif
	tmTimerRoutine=NULL;
	tmResetTimer();
#ifdef TIMER_DEBUG
	tmInited=0;
#endif
}

void tmSetNewRate(int timerval)
{
	struct itimerval spec;

#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmSetNewRate: we are not inited\n");
#endif

	tmTimerRate = timerval;
#ifdef TIMER_DEBUG
	fprintf(stderr, "tmSetNewRate (ticks=%d, time=%ld)\n", timerval, count_to_time(timerval));
#endif

	timerval = count_to_time(timerval);
  
	spec.it_interval.tv_sec=0;
	spec.it_interval.tv_usec=timerval;
	spec.it_value.tv_sec=0;
	spec.it_value.tv_usec=timerval;
	setitimer(ITIMER_REAL, &spec, NULL);
}

int tmGetTimer(void)
{
	unsigned long tm = tmTimerRate+tmTicker;

	unsigned long tv2;
	struct itimerval spec;

#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmGetTimer: we are not inited\n");
#endif

	getitimer(ITIMER_REAL, &spec);
	tv2 = time_to_count(spec.it_value.tv_usec);

	tm -= tv2;
	return umulshr16(tm, 3600);
}

long tmGetTicker(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmGetTicker: we are not inited\n");
#endif
	return tmTicker;
}

void tmSetTicker(long t)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmSetTicker: we are not inited\n");
#endif
	tmTicker+=t-tmGetTicker();
}

int tmGetCpuUsage()
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmGetCpuUsage: we are not inited\n");
#endif
	return (char)cpuusage;
}

void tmSetSecure()
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmSecure: we are not inited\n");
#endif
	secure++;
}

void tmReleaseSecure(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmReleaseSecure: we are not inited\n");
#endif
	if (secure)
		secure--;
}


int pollInit(void (*proc)(void))
{
	tmTimerRoutineSlave=proc;
	if (!tmTimerRoutine) /* If no timer active, just start to count */
		tmInit(NULL, 17100);
	return 1;
}

void pollClose(void)
{
	tmTimerRoutineSlave=NULL;
	tmResetTimer();
}

==> ocp-0.1.6.tar.bz2: ocp-0.1.6/stuff/timer.c <==
/* OpenCP Module Player
 * copyright (c) '94-'05 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *
 * Systemtimer handlers
 *
 * revision history: (please note changes here)
 *  -nb980510   Niklas Beisert <nbeisert@physik.tu-muenchen.de>
 *    -first release
 *  -doj990328  Dirk Jagdmann  <doj@cubic.org>
 *    -changed interrupt access calls to calls from irq.h
 *  -fd990518   Felix Domke  <tmbinc@gmx.net>
 *    -added CLD after the tmOldTimer-call.
 *     this removed the devwmix*-STRANGEBUG. (finally, i hope)
 *  -fd990817   Felix Domke  <tmbinc@gmx.net>
 *    -added tmSetSecure/tmReleaseSecure to ensure that timer is only
 *     called when not "indos". needed for devpVXD (and some other maybe).
 *  -ss040613   Stian Skjelstad <stian@nixia.no>
 *    -Rewritten to use the posix itimer, and posix signals to fetch them
 *  --ss040907  Stian Skjelstad <stian@nixia.no>
 *    -Use gettimeofday() to calculate cpu-usage, since itimer() uses rounded off values
 */

#include "config.h"
#include <signal.h>
#include <stdio.h>
#include <sys/time.h>
#include "irq.h"
#include "imsrtns.h"
#include "timer.h"
#include "poll.h"

#define CLOCK_TICK_RATE 1193180 /* Base hertz of the i8253 or what ever that chip was namned - Stian */

static unsigned long count_to_time(unsigned long ticks)
{
	unsigned long retval;

	if (ticks&0xfffff000)
	{
		ticks*=62500; /* this is safe, since counter on the chip is 16 bit */
		retval = ticks / CLOCK_TICK_RATE;
		retval*=16;
	} else {
		ticks*=1000000;
		retval = ticks / CLOCK_TICK_RATE;
	}
	return retval;
}

static unsigned long time_to_count(unsigned long time)
{
	unsigned long retval;
	
	if (time<=4000)
	{
		retval = time * CLOCK_TICK_RATE / 1000000;
	} else {
		time/=4000;
		retval = time * CLOCK_TICK_RATE / 2500;
	}
	return retval;
}

static void (*tmTimerRoutine)()=NULL;
static void (*tmTimerRoutineSlave)()=NULL;

static unsigned long tmTimerRate;
static unsigned long tmIntCount;
static unsigned long tmTicker;
#ifdef TIMER_DEBUG
static int tmInited = 0;
#endif

static volatile char overload;
static volatile float cpuusage;
static int secure=0;

static int stackused; /* we don't provide a stack, but is for locking */

static void tmTimerHandler(int ignore)
{
#if 0
	char fpubuf[108]; /* 108 is enough acording to my intel book */
	__asm__ __volatile__ ("fsave %0\n" : : "m"(*fpubuf));
#endif

	struct timeval pre, pro;

	gettimeofday(&pre, NULL);
	tmTicker+=tmTimerRate;

	tmIntCount+=tmTimerRate;
	if (tmIntCount&0xFFFFc000)
	{
		tmIntCount&=0x3FFF;
		if (tmTimerRoutineSlave)
			tmTimerRoutineSlave();
	}

	if (stackused++)
	{
		stackused--;
		cpuusage=100;
		overload=1;
		return;
	}
	if (!secure)
	{
		if (tmTimerRoutine)
			tmTimerRoutine();
	}
	stackused--;

	if (!overload)
	{
		struct itimerval spec;
		unsigned long spent;

		getitimer(ITIMER_REAL, &spec);
		gettimeofday(&pro, NULL);
		spent=(pro.tv_sec-pre.tv_sec)*1000000+pro.tv_usec-pre.tv_usec;
		cpuusage=0.1*(100.0*spent/spec.it_interval.tv_usec)+0.9*cpuusage;
	} else
		cpuusage=100;
	overload=0;
#if 0
	__asm__ __volatile__ ("frstor %0\n" : : "m"(*fpubuf));
#endif
}

int tmInit(void (*rout)(), int timerval)
{
	struct itimerval spec;

#ifdef TIMER_DEBUG
	if (tmInited)
	{
		fprintf(stderr, "tmInit: we are already inited\n");
	} else {
		tmInited=1;
	}
#endif

	tmTimerRate = timerval;
	tmTicker = -timerval;
#ifdef TIMER_DEBUG
	fprintf(stderr, "tmInit (ticks=%d, time=%ld)\n", timerval, count_to_time(timerval));
#endif
	timerval = count_to_time(timerval);
  
	tmTimerRoutine=rout;
	tmIntCount=0;

	irqInit(SIGALRM, tmTimerHandler, 1);
	spec.it_interval.tv_sec=0;
	spec.it_interval.tv_usec=timerval;
	spec.it_value.tv_sec=0;
	spec.it_value.tv_usec=timerval;
	setitimer(ITIMER_REAL, &spec, NULL);

	cpuusage=0;

	return 1;
}

void tmResetTimer(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmResetTimer: we are not inited\n");
#endif
	if (!(tmTimerRoutine&&tmTimerRoutineSlave))
	{
		/* shut down irq */
		struct itimerval spec;

		spec.it_interval.tv_sec=0;
		spec.it_interval.tv_usec=0;
		spec.it_value.tv_sec=0;
		spec.it_value.tv_usec=0;
		setitimer(ITIMER_REAL, &spec, NULL);
 
		irqDone(SIGALRM);
	} else if (!tmTimerRoutine)
		/* recalc to a less/more aggressive clock */
		tmSetNewRate(17100);

}

void tmClose(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmClose: we are not inited\n");
#endif
	tmTimerRoutine=NULL;
	tmResetTimer();
#ifdef TIMER_DEBUG
	tmInited=0;
#endif
}

void tmSetNewRate(int timerval)
{
	struct itimerval spec;

#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmSetNewRate: we are not inited\n");
#endif

	tmTimerRate = timerval;
#ifdef TIMER_DEBUG
	fprintf(stderr, "tmSetNewRate (ticks=%d, time=%ld)\n", timerval, count_to_time(timerval));
#endif

	timerval = count_to_time(timerval);
  
	spec.it_interval.tv_sec=0;
	spec.it_interval.tv_usec=timerval;
	spec.it_value.tv_sec=0;
	spec.it_value.tv_usec=timerval;
	setitimer(ITIMER_REAL, &spec, NULL);
}

int tmGetTimer(void)
{
	unsigned long tm = tmTimerRate+tmTicker;

	unsigned long tv2;
	struct itimerval spec;

#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmGetTimer: we are not inited\n");
#endif

	getitimer(ITIMER_REAL, &spec);
	tv2 = time_to_count(spec.it_value.tv_usec);

	tm -= tv2;
	return umulshr16(tm, 3600);
}

long tmGetTicker(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmGetTicker: we are not inited\n");
#endif
	return tmTicker;
}

void tmSetTicker(long t)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmSetTicker: we are not inited\n");
#endif
	tmTicker+=t-tmGetTicker();
}

int tmGetCpuUsage()
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmGetCpuUsage: we are not inited\n");
#endif
	return (char)cpuusage;
}

void tmSetSecure()
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmSecure: we are not inited\n");
#endif
	secure++;
}

void tmReleaseSecure(void)
{
#ifdef TIMER_DEBUG
	if (!tmInited)
		fprintf(stderr, "tmReleaseSecure: we are not inited\n");
#endif
	if (secure)
		secure--;
}


int pollInit(void (*proc)(void))
{
	tmTimerRoutineSlave=proc;
	if (!tmTimerRoutine) /* If no timer active, just start to count */
		tmInit(NULL, 17100);
	return 1;
}

void pollClose(void)
{
	tmTimerRoutineSlave=NULL;
	tmResetTimer();
}
